@@ -1,8 +1,10 @@
 /// document.rs - has Document, for opening, editing and saving documents
-use crate::event::{Error, Event, Result, Status, EventMgmt};
-use crate::map::{CharMap, form_map};
-use crate::searching::{Searcher, Match};
-use crate::utils::{Loc, Size, get_range, trim, width, tab_boundaries_backward, tab_boundaries_forward};
+use crate::event::{Error, Event, EventMgmt, Result, Status};
+use crate::map::{form_map, CharMap};
+use crate::searching::{Match, Searcher};
+use crate::utils::{
+    get_range, tab_boundaries_backward, tab_boundaries_forward, trim, width, Loc, Size,
+};
 use ropey::Rope;
 use std::fs::File;
 use std::io::{BufReader, BufWriter};
@@ -86,7 +88,10 @@ impl Document {
         let file_name = file_name.into();
         let file = Rope::from_reader(BufReader::new(File::open(&file_name)?))?;
         Ok(Self {
-            eol: !file.line(file.len_lines().saturating_sub(1)).to_string().is_empty(),
+            eol: !file
+                .line(file.len_lines().saturating_sub(1))
+                .to_string()
+                .is_empty(),
             file,
             lines: vec![],
             dbl_map: CharMap::default(),
@@ -119,7 +124,8 @@ impl Document {
         if !self.read_only {
             self.modified = false;
             if let Some(file_name) = &self.file_name {
-                self.file.write_to(BufWriter::new(File::create(file_name)?))?;
+                self.file
+                    .write_to(BufWriter::new(File::create(file_name)?))?;
                 Ok(())
             } else {
                 Err(Error::NoFileName)
@@ -135,7 +141,8 @@ impl Document {
     /// or character set issues.
     pub fn save_as(&self, file_name: &str) -> Result<()> {
         if !self.read_only {
-            self.file.write_to(BufWriter::new(File::create(file_name)?))?;
+            self.file
+                .write_to(BufWriter::new(File::create(file_name)?))?;
             Ok(())
         } else {
             Err(Error::ReadOnlyFile)
@@ -229,7 +236,7 @@ impl Document {
         self.tab_map.splice(loc, tab_start, tabs);
         // Go to end x position
         self.move_to_x(loc.x + st.chars().count());
-        self.old_cursor = self.char_ptr;
+        self.old_cursor = self.loc().x;
         Ok(())
     }
 
@@ -237,15 +244,15 @@ impl Document {
     pub fn delete_with_tab(&mut self, loc: &Loc, st: &str) -> Result<()> {
         // Check for tab spaces
         let boundaries = tab_boundaries_backward(
-            &self.line(loc.y).unwrap_or_else(|| "".to_string()), 
-            self.tab_width
+            &self.line(loc.y).unwrap_or_else(|| "".to_string()),
+            self.tab_width,
         );
         if boundaries.contains(&loc.x.saturating_add(1)) && !self.in_redo {
             // Register other delete actions to delete the whole tab
             let mut loc_copy = loc.clone();
             self.delete(loc.x..=loc.x + st.chars().count(), loc.y)?;
             for _ in 1..self.tab_width {
-                loc_copy.x -= 1;
+                loc_copy.x = loc_copy.x.saturating_sub(1);
                 self.exe(Event::Delete(loc_copy, " ".to_string()))?;
             }
             Ok(())
@@ -273,14 +280,24 @@ impl Document {
         end += line_start;
         let removed = self.file.slice(start..end).to_string();
         // Update unicode and tab map
-        self.dbl_map.shift_deletion(&Loc::at(line_start, y), (start, end), &removed, self.tab_width);
-        self.tab_map.shift_deletion(&Loc::at(line_start, y), (start, end), &removed, self.tab_width);
+        self.dbl_map.shift_deletion(
+            &Loc::at(line_start, y),
+            (start, end),
+            &removed,
+            self.tab_width,
+        );
+        self.tab_map.shift_deletion(
+            &Loc::at(line_start, y),
+            (start, end),
+            &removed,
+            self.tab_width,
+        );
         // Update rope
         self.file.remove(start..end);
         // Update cache
         let line: String = self.file.line(y).chars().collect();
         self.lines[y] = line.trim_end_matches(&['\n', '\r']).to_string();
-        self.old_cursor = self.char_ptr;
+        self.old_cursor = self.loc().x;
         Ok(())
     }
 
@@ -309,7 +326,7 @@ impl Document {
         self.loaded_to += 1;
         // Goto line
         self.move_to_y(loc);
-        self.old_cursor = self.char_ptr;
+        self.old_cursor = self.loc().x;
         Ok(())
     }
 
@@ -334,7 +351,7 @@ impl Document {
         self.loaded_to = self.loaded_to.saturating_sub(1);
         // Goto line
         self.move_to_y(loc);
-        self.old_cursor = self.char_ptr;
+        self.old_cursor = self.loc().x;
         Ok(())
     }
 
@@ -351,7 +368,7 @@ impl Document {
         self.delete(loc.x.., loc.y)?;
         self.insert_line(loc.y + 1, rhs)?;
         self.move_to(&Loc::at(0, loc.y + 1));
-        self.old_cursor = self.char_ptr;
+        self.old_cursor = self.loc().x;
         Ok(())
     }
 
@@ -368,7 +385,7 @@ impl Document {
         self.delete_line(y + 1)?;
         self.insert(&Loc::at(length, y), &below)?;
         self.move_to(&Loc::at(length, y));
-        self.old_cursor = self.char_ptr;
+        self.old_cursor = self.loc().x;
         Ok(())
     }
 
@@ -402,7 +419,8 @@ impl Document {
         if self.loc().y == 0 {
             return Status::StartOfFile;
         }
-        self.cursor.loc.y -= 1;
+        self.cursor.loc.y = self.cursor.loc.y.saturating_sub(1);
+        self.cursor.loc.x = self.old_cursor;
         // Snap to end of line
         self.fix_dangling_cursor();
         // Move back if in the middle of a longer character
@@ -427,6 +445,7 @@ impl Document {
             return Status::EndOfFile;
         }
         self.cursor.loc.y += 1;
+        self.cursor.loc.x = self.old_cursor;
         // Snap to end of line
         self.fix_dangling_cursor();
         // Move back if in the middle of a longer character
@@ -451,12 +470,25 @@ impl Document {
             return Status::StartOfLine;
         }
         // Determine the width of the character to traverse
-        let width = self.width_of(self.loc().y, self.char_ptr.saturating_sub(1));
+        let line = self.line(self.loc().y).unwrap_or_else(|| "".to_string());
+        let boundaries = tab_boundaries_backward(&line, self.tab_width);
+        let width = if boundaries.contains(&self.char_ptr) {
+            // Push the character pointer up
+            self.char_ptr = self
+                .char_ptr
+                .saturating_sub(self.tab_width.saturating_sub(1));
+            // There are spaces that should be treated as tabs (so should traverse the tab width)
+            self.tab_width
+        } else {
+            // There are no spaces that should be treated as tabs
+            self.width_of(self.loc().y, self.char_ptr.saturating_sub(1))
+        };
         // Move back the correct amount
-        self.cursor.loc.x -= width;
+        self.cursor.loc.x = self.cursor.loc.x.saturating_sub(width);
         // Update the character pointer
-        self.char_ptr -= 1;
+        self.char_ptr = self.char_ptr.saturating_sub(1);
         self.bring_cursor_in_viewport();
+        self.old_cursor = self.loc().x;
         Status::None
     }
 
@@ -491,7 +523,7 @@ impl Document {
         // Update the character pointer
         self.char_ptr += 1;
         self.bring_cursor_in_viewport();
-        self.old_cursor = self.char_ptr;
+        self.old_cursor = self.loc().x;
         Status::None
     }
 
@@ -505,6 +537,7 @@ impl Document {
     pub fn select_home(&mut self) {
         self.cursor.loc.x = 0;
         self.char_ptr = 0;
+        self.old_cursor = 0;
         self.bring_cursor_in_viewport();
     }
 
@@ -519,6 +552,7 @@ impl Document {
         let line = self.line(self.loc().y).unwrap_or_else(|| "".to_string());
         let length = line.chars().count();
         self.select_to_x(length);
+        self.old_cursor = self.loc().x;
     }
 
     /// Move to the top of the document
@@ -535,22 +569,26 @@ impl Document {
     /// Select to the top of the document
     pub fn select_top(&mut self) {
         self.select_to(&Loc::at(0, 0));
+        self.old_cursor = self.loc().x;
     }
 
     /// Select to the bottom of the document
     pub fn select_bottom(&mut self) {
         let last = self.len_lines();
         self.select_to(&Loc::at(0, last));
+        self.old_cursor = self.loc().x;
     }
 
     /// Move up by 1 page
     pub fn move_page_up(&mut self) {
         self.move_to_y(self.cursor.loc.y.saturating_sub(self.size.h));
+        self.old_cursor = 0;
     }
 
     /// Move down by 1 page
     pub fn move_page_down(&mut self) {
         self.move_to_y(self.cursor.loc.y + self.size.h);
+        self.old_cursor = 0;
     }
 
     /// Moves to the previous word in the document
@@ -571,7 +609,7 @@ impl Document {
                 return self.move_prev_word();
             }
         }
-        self.old_cursor = self.char_ptr;
+        self.old_cursor = self.loc().x;
         Status::None
     }
 
@@ -587,7 +625,7 @@ impl Document {
             mtch.loc.x += mtch.text.chars().count();
             self.move_to(&mtch.loc);
         }
-        self.old_cursor = self.char_ptr;
+        self.old_cursor = self.loc().x;
         Status::None
     }
 
@@ -596,14 +634,15 @@ impl Document {
         // Prepare
         let mut srch = Searcher::new(regex);
         // Check current line for matches
-        let current: String = self.line(self.loc().y)?
+        let current: String = self
+            .line(self.loc().y)?
             .chars()
             .skip(self.char_ptr + inc)
             .collect();
         if let Some(mut mtch) = srch.lfind(&current) {
             mtch.loc.y = self.loc().y;
             mtch.loc.x += self.char_ptr + inc;
-            return Some(mtch)
+            return Some(mtch);
         }
         // Check subsequent lines for matches
         let mut line_no = self.loc().y + 1;
@@ -624,7 +663,8 @@ impl Document {
         // Prepare
         let mut srch = Searcher::new(regex);
         // Check current line for matches
-        let current: String = self.line(self.loc().y)?
+        let current: String = self
+            .line(self.loc().y)?
             .chars()
             .take(self.char_ptr)
             .collect();
@@ -640,7 +680,9 @@ impl Document {
                 mtch.loc.y = line_no;
                 return Some(mtch);
             }
-            if line_no == 0 { break; }
+            if line_no == 0 {
+                break;
+            }
             line_no = line_no.saturating_sub(1);
         }
         None
@@ -690,7 +732,9 @@ impl Document {
         }
         // If the move position is out of bounds, move to the end of the line
         if line.chars().count() < x {
-            self.select_end();
+            let line = self.line(self.loc().y).unwrap_or_else(|| "".to_string());
+            let length = line.chars().count();
+            self.select_to_x(length);
             return;
         }
         // Update char position
@@ -706,7 +750,7 @@ impl Document {
     pub fn move_to_y(&mut self, y: usize) {
         self.select_to_y(y);
         self.cancel_selection();
-    }    
+    }
 
     /// Function to select to a specific y position
     pub fn select_to_y(&mut self, y: usize) {
@@ -731,13 +775,13 @@ impl Document {
             self.offset.y = self.cursor.loc.y;
         }
         if self.offset.y + self.size.h <= self.cursor.loc.y {
-            self.offset.y = self.cursor.loc.y - self.size.h + 1;
+            self.offset.y = self.cursor.loc.y.saturating_sub(self.size.h) + 1;
         }
         if self.offset.x > self.cursor.loc.x {
             self.offset.x = self.cursor.loc.x;
         }
         if self.offset.x + self.size.w <= self.cursor.loc.x {
-            self.offset.x = self.cursor.loc.x - self.size.w + 1;
+            self.offset.x = self.cursor.loc.x.saturating_sub(self.size.w) + 1;
         }
         self.load_to(self.offset.y + self.size.h);
     }
@@ -769,9 +813,11 @@ impl Document {
     pub fn character_idx(&self, loc: &Loc) -> usize {
         let mut idx = loc.x;
         // Account for double width characters
-        idx -= self.dbl_map.count(loc, true).unwrap_or(0);
+        idx = idx.saturating_sub(self.dbl_map.count(loc, true).unwrap_or(0));
         // Account for tab characters
-        idx -= self.tab_map.count(loc, true).unwrap_or(0) * self.tab_width.saturating_sub(1);
+        idx = idx.saturating_sub(
+            self.tab_map.count(loc, true).unwrap_or(0) * self.tab_width.saturating_sub(1),
+        );
         idx
     }
 
@@ -789,9 +835,9 @@ impl Document {
     fn update_char_ptr(&mut self) {
         let mut idx = self.loc().x;
         let dbl_count = self.dbl_map.count(&self.loc(), true).unwrap_or(0);
-        idx -= dbl_count;
+        idx = idx.saturating_sub(dbl_count);
         let tab_count = self.tab_map.count(&self.loc(), true).unwrap_or(0);
-        idx -= tab_count * self.tab_width.saturating_sub(1);
+        idx = idx.saturating_sub(tab_count * self.tab_width.saturating_sub(1));
         self.char_ptr = idx;
     }
 
@@ -811,21 +857,29 @@ impl Document {
         let mut magnitude = 0;
         let Loc { x, y } = self.loc();
         if let Some(map) = self.dbl_map.get(y) {
-            let last_dbl = self.dbl_map.count(&self.loc(), true).unwrap().saturating_sub(1);
+            let last_dbl = self
+                .dbl_map
+                .count(&self.loc(), true)
+                .unwrap()
+                .saturating_sub(1);
             let start = map[last_dbl].0;
             if x == start + 1 {
                 magnitude += 1;
             }
         }
         if let Some(map) = self.tab_map.get(y) {
-            let last_tab = self.tab_map.count(&self.loc(), true).unwrap().saturating_sub(1);
+            let last_tab = self
+                .tab_map
+                .count(&self.loc(), true)
+                .unwrap()
+                .saturating_sub(1);
             let start = map[last_tab].0;
             let range = start..start + self.tab_width;
             if range.contains(&x) {
-                magnitude += x - start;
+                magnitude += x.saturating_sub(start);
             }
         }
-        self.cursor.loc.x -= magnitude;
+        self.cursor.loc.x = self.cursor.loc.x.saturating_sub(magnitude);
     }
 
     /// Load lines in this document up to a specified index.
@@ -847,7 +901,8 @@ impl Document {
                 self.dbl_map.insert(i, dbl_map);
                 self.tab_map.insert(i, tab_map);
                 // Cache this line
-                self.lines.push(line.trim_end_matches(&['\n', '\r']).to_string());
+                self.lines
+                    .push(line.trim_end_matches(&['\n', '\r']).to_string());
             }
             // Store new loaded point
             self.loaded_to = to;
@@ -882,7 +937,7 @@ impl Document {
         } else {
             (request + 1).to_string()
         };
-        format!("{}{}", " ".repeat(total - num.len()), num)
+        format!("{}{}", " ".repeat(total.saturating_sub(num.len())), num)
     }
 
     /// Determine if a character at a certain location is a double width character.
@@ -946,8 +1001,8 @@ impl Document {
             return None;
         }
         let result = Loc {
-            x: self.cursor.loc.x - self.offset.x,
-            y: self.cursor.loc.y - self.offset.y,
+            x: self.cursor.loc.x.saturating_sub(self.offset.x),
+            y: self.cursor.loc.y.saturating_sub(self.offset.y),
         };
         if result.x > self.size.w || result.y > self.size.h {
             return None;
@@ -976,7 +1031,7 @@ impl Document {
     /// Returns true if the provided location is within the current active selection
     pub fn is_loc_selected(&self, loc: Loc) -> bool {
         let (left, right) = self.selection_loc_bound();
-        left <= loc && loc < right 
+        left <= loc && loc < right
     }
 
     /// Will return the current active selection as a range over file characters
