@@ -1,18 +1,18 @@
-use crate::ui::{size, Terminal, Feedback, HELP_TEXT};
-use crate::config::{Config, key_to_string};
+use crate::config::{key_to_string, Config};
 use crate::error::{OxError, Result};
+use crate::ui::{size, Feedback, Terminal, HELP_TEXT};
 use crossterm::{
     event::{read, Event as CEvent, KeyCode as KCode, KeyModifiers as KMod},
-    style::{SetBackgroundColor as Bg, SetForegroundColor as Fg, SetAttribute, Attribute},
+    style::{Attribute, SetAttribute, SetBackgroundColor as Bg, SetForegroundColor as Fg},
     terminal::{Clear, ClearType as ClType},
 };
-use kaolinite::event::{Event, Status, Error as KError};
+use kaolinite::event::{Error as KError, Event, Status};
 use kaolinite::utils::{Loc, Size};
 use kaolinite::Document;
-use synoptic::{Highlighter, TokOpt, trim};
-use std::time::Instant;
-use std::io::{Write, ErrorKind};
 use mlua::Lua;
+use std::io::{ErrorKind, Write};
+use std::time::Instant;
+use synoptic::{trim, Highlighter, TokOpt};
 
 mod mouse;
 
@@ -42,6 +42,10 @@ pub struct Editor {
     pub command: Option<String>,
     /// Will store the last time the editor was interacted with (to track inactivity)
     pub last_active: Instant,
+    /// Used for storing amount to push document down
+    push_down: usize,
+    /// Used to cache the location of the configuration file
+    pub config_path: String,
 }
 
 impl Editor {
@@ -61,20 +65,39 @@ impl Editor {
             feedback: Feedback::None,
             command: None,
             last_active: Instant::now(),
+            push_down: 1,
+            config_path: "~/.oxrc".to_string(),
         })
     }
 
     /// Load the configuration values
     pub fn load_config(&mut self, path: String, lua: &Lua) -> Result<()> {
-        self.config.read(path, lua)?;
+        self.config_path = path.clone();
+        let result = self.config.read(path, lua);
+        // Display any warnings if the user configuration couldn't be found
+        if let Err(OxError::Config(msg)) = result {
+            if msg == "Not Found" {
+                let warn = "No configuration file found, using default configuration".to_string();
+                self.feedback = Feedback::Warning(warn);
+            }
+        } else {
+            result?
+        };
+        // Calculate the correct push down based on config
+        self.push_down = if self.config.tab_line.borrow().enabled {
+            1
+        } else {
+            0
+        };
         Ok(())
     }
 
     /// Function to create a new document
     pub fn blank(&mut self) -> Result<()> {
         let mut size = size()?;
-        size.h = size.h.saturating_sub(2);
+        size.h = size.h.saturating_sub(1 + self.push_down);
         let mut doc = Document::new(size);
+        doc.set_tab_width(self.config.document.borrow().tab_width);
         // Load all the lines within viewport into the document
         doc.load_to(size.h);
         // Update in the syntax highlighter
@@ -96,13 +119,18 @@ impl Editor {
     /// Function to open a document into the editor
     pub fn open(&mut self, file_name: String) -> Result<()> {
         let mut size = size()?;
-        size.h -= 2;
+        size.h = size.h.saturating_sub(1 + self.push_down);
         let mut doc = Document::open(size, file_name.clone())?;
+        doc.set_tab_width(self.config.document.borrow().tab_width);
         // Load all the lines within viewport into the document
         doc.load_to(size.h);
         // Update in the syntax highlighter
-        let ext = file_name.split('.').last().unwrap();
-        let mut highlighter = self.config
+        let mut ext = file_name.split('.').last().unwrap_or("");
+        if ext == "oxrc" {
+            ext = "lua"
+        }
+        let mut highlighter = self
+            .config
             .syntax_highlighting
             .borrow()
             .get_highlighter(&ext);
@@ -128,15 +156,19 @@ impl Editor {
             if os.kind() == ErrorKind::NotFound {
                 self.blank()?;
                 let binding = file_name.clone();
-                let ext = binding.split('.').last().unwrap();
+                let ext = binding.split('.').last().unwrap_or("");
                 self.doc.last_mut().unwrap().file_name = Some(file_name);
                 self.doc.last_mut().unwrap().modified = true;
-                let highlighter = self.config
+                let highlighter = self
+                    .config
                     .syntax_highlighting
                     .borrow()
                     .get_highlighter(&ext);
                 *self.highlighter.last_mut().unwrap() = highlighter;
-                self.highlighter.last_mut().unwrap().run(&self.doc.last_mut().unwrap().lines);
+                self.highlighter
+                    .last_mut()
+                    .unwrap()
+                    .run(&self.doc.last_mut().unwrap().lines);
                 Ok(())
             } else {
                 file
@@ -201,9 +233,9 @@ impl Editor {
 
     /// Complete one cycle of the editor
     /// This function will return a key press code if applicable
-    pub fn cycle(&mut self) -> Result<Option<String>> {
+    pub fn cycle(&mut self, lua: &Lua) -> Result<Option<String>> {
         // Run the editor
-        self.render()?;
+        self.render(&lua)?;
         // Wait for an event
         let event = read()?;
         self.needs_rerender = match event {
@@ -218,7 +250,7 @@ impl Editor {
                 // Check period of inactivity and commit events (for undo/redo) if over 10secs
                 let end = Instant::now();
                 let inactivity = end.duration_since(self.last_active).as_secs() as usize;
-                if inactivity > 10 {
+                if inactivity > self.config.document.borrow().undo_period {
                     self.doc_mut().event_mgmt.commit();
                 }
                 self.last_active = Instant::now();
@@ -240,7 +272,7 @@ impl Editor {
                 self.doc_mut().size.w = w.saturating_sub(max as u16) as usize;
                 self.doc_mut().size.h = h.saturating_sub(3) as usize;
                 let max = self.doc().offset.x + self.doc().size.h;
-                self.doc_mut().load_to(max);
+                self.doc_mut().load_to(max + 1);
             }
             CEvent::Mouse(mouse_event) => {
                 self.handle_mouse_event(mouse_event);
@@ -255,10 +287,14 @@ impl Editor {
     /// Append any missed lines to the syntax highlighter
     pub fn update_highlighter(&mut self) -> Result<()> {
         if self.active {
-            let actual = self.doc.get(self.ptr).and_then(|d| Some(d.loaded_to)).unwrap_or(0);
+            let actual = self
+                .doc
+                .get(self.ptr)
+                .and_then(|d| Some(d.loaded_to))
+                .unwrap_or(0);
             let percieved = self.highlighter().line_ref.len();
             if percieved < actual {
-                let diff = actual - percieved;
+                let diff = actual.saturating_sub(percieved);
                 for i in 0..diff {
                     let line = &self.doc[self.ptr].lines[percieved + i];
                     self.highlighter[self.ptr].append(line);
@@ -269,23 +305,26 @@ impl Editor {
     }
 
     /// Render a single frame of the editor in it's current state
-    pub fn render(&mut self) -> Result<()> {
+    pub fn render(&mut self, lua: &Lua) -> Result<()> {
         if !self.needs_rerender {
             return Ok(());
         }
         self.needs_rerender = false;
         self.terminal.hide_cursor()?;
         let Size { w, mut h } = size()?;
-        h = h.saturating_sub(2);
+        h = h.saturating_sub(1 + self.push_down);
         // Update the width of the document in case of update
         let max = self.dent();
         self.doc_mut().size.w = w.saturating_sub(max) as usize;
         // Render the tab line
-        self.render_tab_line(w)?;
+        let tab_enabled = self.config.tab_line.borrow().enabled;
+        if tab_enabled {
+            self.render_tab_line(w)?;
+        }
         // Run through each line of the terminal, rendering the correct line
         self.render_document(w, h)?;
         // Leave last line for status line
-        self.render_status_line(w, h)?;
+        self.render_status_line(&lua, w, h)?;
         // Render greeting or help message if applicable
         if self.greet {
             self.render_greeting(w, h)?;
@@ -297,7 +336,7 @@ impl Editor {
         // Move cursor to the correct location and perform render
         if let Some(Loc { x, y }) = self.doc().cursor_loc_in_screen() {
             self.terminal.show_cursor()?;
-            self.terminal.goto(x + max, y + 1)?;
+            self.terminal.goto(x + max, y + self.push_down)?;
         }
         self.terminal.flush()?;
         Ok(())
@@ -306,19 +345,32 @@ impl Editor {
     /// Render the lines of the document
     fn render_document(&mut self, _w: usize, h: usize) -> Result<()> {
         for y in 0..(h as u16) {
-            self.terminal.goto(0, y + 1)?;
+            self.terminal
+                .goto(0, y as usize + self.push_down as usize)?;
             // Start background colour
-            write!(self.terminal.stdout, "{}", Bg(self.config.colors.borrow().editor_bg.to_color()?))?;
-            write!(self.terminal.stdout, "{}", Fg(self.config.colors.borrow().editor_fg.to_color()?))?;
+            write!(
+                self.terminal.stdout,
+                "{}",
+                Bg(self.config.colors.borrow().editor_bg.to_color()?)
+            )?;
+            write!(
+                self.terminal.stdout,
+                "{}",
+                Fg(self.config.colors.borrow().editor_fg.to_color()?)
+            )?;
             // Write line number of document
             if self.config.line_numbers.borrow().enabled {
                 let num = self.doc().line_number(y as usize + self.doc().offset.y);
+                let padding_left = " ".repeat(self.config.line_numbers.borrow().padding_left);
+                let padding_right = " ".repeat(self.config.line_numbers.borrow().padding_right);
                 write!(
                     self.terminal.stdout,
-                    "{}{} {} │{}{}",
+                    "{}{}{}{}{}│{}{}",
                     Bg(self.config.colors.borrow().line_number_bg.to_color()?),
                     Fg(self.config.colors.borrow().line_number_fg.to_color()?),
+                    padding_left,
                     num,
+                    padding_right,
                     Fg(self.config.colors.borrow().editor_fg.to_color()?),
                     Bg(self.config.colors.borrow().editor_bg.to_color()?),
                 )?;
@@ -338,12 +390,8 @@ impl Editor {
                             match colour {
                                 // Success, write token
                                 Ok(col) => {
-                                    write!(
-                                        self.terminal.stdout,
-                                        "{}",
-                                        Fg(col),
-                                    )?;
-                                },
+                                    write!(self.terminal.stdout, "{}", Fg(col),)?;
+                                }
                                 // Failure, show error message and don't highlight this token
                                 Err(err) => {
                                     self.feedback = Feedback::Error(err.to_string());
@@ -356,31 +404,40 @@ impl Editor {
                     for c in text.chars() {
                         let is_selected = self.doc().is_loc_selected(Loc { y: idx, x: x_pos });
                         if is_selected {
-                            write!(self.terminal.stdout, "{}", Bg(self.config.colors.borrow().selection_bg.to_color()?))?;
-                            write!(self.terminal.stdout, "{}", Fg(self.config.colors.borrow().selection_fg.to_color()?))?;
+                            write!(
+                                self.terminal.stdout,
+                                "{}",
+                                Bg(self.config.colors.borrow().selection_bg.to_color()?)
+                            )?;
+                            write!(
+                                self.terminal.stdout,
+                                "{}",
+                                Fg(self.config.colors.borrow().selection_fg.to_color()?)
+                            )?;
                         } else {
-                            write!(self.terminal.stdout, "{}", Bg(self.config.colors.borrow().editor_bg.to_color()?))?;
+                            write!(
+                                self.terminal.stdout,
+                                "{}",
+                                Bg(self.config.colors.borrow().editor_bg.to_color()?)
+                            )?;
                         }
                         write!(self.terminal.stdout, "{c}")?;
                         x_pos += 1;
                     }
-                    write!(self.terminal.stdout, "{}", Fg(self.config.colors.borrow().editor_fg.to_color()?))?;
+                    write!(
+                        self.terminal.stdout,
+                        "{}",
+                        Fg(self.config.colors.borrow().editor_fg.to_color()?)
+                    )?;
                 }
             }
         }
         Ok(())
     }
 
-    /// Put together the contents of a tab
-    fn render_document_tab_header(&self, document: &Document) -> String {
-        let file_name = document.file_name.clone().unwrap_or_else(|| "[No Name]".to_string());
-        let modified = if document.modified { "[+]" } else { "" };
-        format!("  {file_name}{modified}  ")
-    }
-
     /// Render the tab line at the top of the document
     fn render_tab_line(&mut self, w: usize) -> Result<()> {
-        self.terminal.prepare_line(0)?;
+        self.terminal.goto(0 as usize, 0 as usize)?;
         write!(
             self.terminal.stdout,
             "{}{}",
@@ -388,7 +445,7 @@ impl Editor {
             Bg(self.config.colors.borrow().tab_inactive_bg.to_color()?)
         )?;
         for (c, document) in self.doc.iter().enumerate() {
-            let document_header = self.render_document_tab_header(document);
+            let document_header = self.config.tab_line.borrow().render(document);
             if c == self.ptr {
                 // Representing the document we're currently looking at
                 write!(
@@ -411,15 +468,15 @@ impl Editor {
     }
 
     /// Render the status line at the bottom of the document
-    fn render_status_line(&mut self, w: usize, h: usize) -> Result<()> {
-        self.terminal.goto(0, h + 1)?;
+    fn render_status_line(&mut self, lua: &Lua, w: usize, h: usize) -> Result<()> {
+        self.terminal.goto(0, h + self.push_down)?;
         write!(
             self.terminal.stdout,
             "{}{}{}{}{}{}{}",
             Bg(self.config.colors.borrow().status_bg.to_color()?),
             Fg(self.config.colors.borrow().status_fg.to_color()?),
             SetAttribute(Attribute::Bold),
-            self.config.status_line.borrow().render(&self, w),
+            self.config.status_line.borrow().render(&self, &lua, w),
             SetAttribute(Attribute::Reset),
             Fg(self.config.colors.borrow().editor_fg.to_color()?),
             Bg(self.config.colors.borrow().editor_bg.to_color()?),
@@ -443,8 +500,8 @@ impl Editor {
         let color = self.config.colors.borrow().highlight.to_color()?;
         let editor_fg = self.config.colors.borrow().editor_fg.to_color()?;
         let message: Vec<&str> = HELP_TEXT.split('\n').collect();
-        for (c, line) in message.iter().enumerate().take(h - h / 4) {
-            self.terminal.goto(w - 30, h / 4 + c + 1)?;
+        for (c, line) in message.iter().enumerate().take(h.saturating_sub(h / 4)) {
+            self.terminal.goto(w.saturating_sub(30), h / 4 + c + 1)?;
             write!(self.terminal.stdout, "{}{line}{}", Fg(color), Fg(editor_fg))?;
         }
         Ok(())
@@ -455,13 +512,12 @@ impl Editor {
         let colors = self.config.colors.borrow();
         let greeting = self.config.greeting_message.borrow().render(&colors)?;
         let message: Vec<&str> = greeting.split('\n').collect();
-        for (c, line) in message.iter().enumerate().take(h - h / 4) {
+        for (c, line) in message.iter().enumerate().take(h.saturating_sub(h / 4)) {
             self.terminal.goto(4, h / 4 + c + 1)?;
             write!(
                 self.terminal.stdout,
                 "{}",
-                alinio::align::center(&line, w - 4)
-                    .unwrap_or_else(|| "".to_string()),
+                alinio::align::center(&line, w.saturating_sub(4)).unwrap_or_else(|| "".to_string()),
             )?;
         }
         Ok(())
@@ -478,8 +534,18 @@ impl Editor {
             let w = size()?.w;
             // Render prompt message
             self.terminal.prepare_line(h)?;
-            write!(self.terminal.stdout, "{}", Bg(self.config.colors.borrow().editor_bg.to_color()?))?;
-            write!(self.terminal.stdout, "{}: {}{}", prompt, input, " ".to_string().repeat(w))?;
+            write!(
+                self.terminal.stdout,
+                "{}",
+                Bg(self.config.colors.borrow().editor_bg.to_color()?)
+            )?;
+            write!(
+                self.terminal.stdout,
+                "{}: {}{}",
+                prompt,
+                input,
+                " ".to_string().repeat(w)
+            )?;
             self.terminal.goto(prompt.len() + input.len() + 2, h)?;
             self.terminal.flush()?;
             // Handle events
@@ -488,7 +554,9 @@ impl Editor {
                     // Exit the menu when the enter key is pressed
                     (KMod::NONE, KCode::Enter) => done = true,
                     // Remove from the input string if the user presses backspace
-                    (KMod::NONE, KCode::Backspace) => { input.pop(); },
+                    (KMod::NONE, KCode::Backspace) => {
+                        input.pop();
+                    }
                     // Add to the input string if the user presses a character
                     (KMod::NONE | KMod::SHIFT, KCode::Char(c)) => input.push(c),
                     _ => (),
@@ -502,7 +570,9 @@ impl Editor {
     /// Work out how much to push the document to the right (to make way for line numbers)
     fn dent(&self) -> usize {
         if self.config.line_numbers.borrow().enabled {
-            self.doc().len_lines().to_string().len() + 3
+            let padding_left = self.config.line_numbers.borrow().padding_left;
+            let padding_right = self.config.line_numbers.borrow().padding_right;
+            self.doc().len_lines().to_string().len() + 1 + padding_left + padding_right
         } else {
             0
         }
@@ -518,7 +588,7 @@ impl Editor {
     /// Move to the previous document opened in the editor
     pub fn prev(&mut self) {
         if self.ptr != 0 {
-            self.ptr -= 1;
+            self.ptr = self.ptr.saturating_sub(1);
         }
     }
 
@@ -531,7 +601,7 @@ impl Editor {
     /// Paste the selected text
     pub fn paste(&mut self) -> Result<()> {
         let clip = self.terminal.paste();
-        for ch in clip.unwrap().chars() {
+        for ch in clip.unwrap_or_else(|| "".to_string()).chars() {
             if let Err(err) = self.character(ch) {
                 self.feedback = Feedback::Error(err.to_string());
             }
@@ -554,7 +624,8 @@ impl Editor {
     pub fn select_left(&mut self) {
         let status = self.doc_mut().select_left();
         // Cursor wrapping if cursor hits the start of the line
-        if status == Status::StartOfLine && self.doc().loc().y != 0 {
+        let wrapping = self.config.document.borrow().wrap_cursor;
+        if status == Status::StartOfLine && self.doc().loc().y != 0 && wrapping {
             self.doc_mut().select_up();
             self.doc_mut().select_end();
         }
@@ -564,7 +635,8 @@ impl Editor {
     pub fn select_right(&mut self) {
         let status = self.doc_mut().select_right();
         // Cursor wrapping if cursor hits the end of a line
-        if status == Status::EndOfLine {
+        let wrapping = self.config.document.borrow().wrap_cursor;
+        if status == Status::EndOfLine && wrapping {
             self.doc_mut().select_down();
             self.doc_mut().select_home();
         }
@@ -590,7 +662,8 @@ impl Editor {
     pub fn left(&mut self) {
         let status = self.doc_mut().move_left();
         // Cursor wrapping if cursor hits the start of the line
-        if status == Status::StartOfLine && self.doc().loc().y != 0 {
+        let wrapping = self.config.document.borrow().wrap_cursor;
+        if status == Status::StartOfLine && self.doc().loc().y != 0 && wrapping {
             self.doc_mut().move_up();
             self.doc_mut().move_end();
         }
@@ -600,7 +673,8 @@ impl Editor {
     pub fn right(&mut self) {
         let status = self.doc_mut().move_right();
         // Cursor wrapping if cursor hits the end of a line
-        if status == Status::EndOfLine {
+        let wrapping = self.config.document.borrow().wrap_cursor;
+        if status == Status::EndOfLine && wrapping {
             self.doc_mut().move_down();
             self.doc_mut().move_home();
         }
@@ -609,7 +683,8 @@ impl Editor {
     /// Move the cursor to the previous word in the line
     pub fn prev_word(&mut self) {
         let status = self.doc_mut().move_prev_word();
-        if status == Status::StartOfLine {
+        let wrapping = self.config.document.borrow().wrap_cursor;
+        if status == Status::StartOfLine && wrapping {
             self.doc_mut().move_up();
             self.doc_mut().move_end();
         }
@@ -618,7 +693,8 @@ impl Editor {
     /// Move the cursor to the next word in the line
     pub fn next_word(&mut self) {
         let status = self.doc_mut().move_next_word();
-        if status == Status::EndOfLine {
+        let wrapping = self.config.document.borrow().wrap_cursor;
+        if status == Status::EndOfLine && wrapping {
             self.doc_mut().move_down();
             self.doc_mut().move_home();
         }
@@ -673,7 +749,7 @@ impl Editor {
             self.new_row()?;
             let mut loc = self.doc().char_loc();
             self.highlighter().remove_line(loc.y);
-            loc.y -= 1;
+            loc.y = loc.y.saturating_sub(1);
             loc.x = self.doc().line(loc.y).unwrap().chars().count();
             self.exe(Event::SpliceUp(loc))?;
             let line = &self.doc[self.ptr].lines[loc.y];
@@ -683,7 +759,10 @@ impl Editor {
             c = c.saturating_sub(1);
             if let Some(line) = self.doc().line(self.doc().loc().y) {
                 if let Some(ch) = line.chars().nth(c) {
-                    let loc = Loc { x: c, y: self.doc().loc().y };
+                    let loc = Loc {
+                        x: c,
+                        y: self.doc().loc().y,
+                    };
                     self.exe(Event::Delete(loc, ch.to_string()))?;
                     self.highlighter[self.ptr].edit(loc.y, &self.doc[self.ptr].lines[loc.y]);
                 }
@@ -697,7 +776,10 @@ impl Editor {
         let c = self.doc().char_ptr;
         if let Some(line) = self.doc().line(self.doc().loc().y) {
             if let Some(ch) = line.chars().nth(c) {
-                let loc = Loc { x: c, y: self.doc().loc().y };
+                let loc = Loc {
+                    x: c,
+                    y: self.doc().loc().y,
+                };
                 self.exe(Event::Delete(loc, ch.to_string()))?;
                 self.highlighter[self.ptr].edit(loc.y, &self.doc[self.ptr].lines[loc.y]);
             }
@@ -740,10 +822,13 @@ impl Editor {
         while !done {
             // Render just the document part
             self.terminal.hide_cursor()?;
-            self.render_document(w, h - 2)?;
+            self.render_document(w, h.saturating_sub(2))?;
             // Render custom status line with mode information
-            self.terminal.prepare_line(h)?;
-            write!(self.terminal.stdout, "[<-]: Search previous | [->]: Search next")?;
+            self.terminal.goto(0, h)?;
+            write!(
+                self.terminal.stdout,
+                "[<-]: Search previous | [->]: Search next"
+            )?;
             self.terminal.flush()?;
             // Move back to correct cursor position
             if let Some(Loc { x, y }) = self.doc().cursor_loc_in_screen() {
@@ -813,10 +898,13 @@ impl Editor {
         while !done {
             // Render just the document part
             self.terminal.hide_cursor()?;
-            self.render_document(w, h - 2)?;
+            self.render_document(w, h.saturating_sub(2))?;
             // Write custom status line for the replace mode
-            self.terminal.prepare_line(h)?;
-            write!(self.terminal.stdout, "[<-] Previous | [->] Next | [Enter] Replace | [Tab] Replace All")?;
+            self.terminal.goto(0, h)?;
+            write!(
+                self.terminal.stdout,
+                "[<-] Previous | [->] Next | [Enter] Replace | [Tab] Replace All"
+            )?;
             self.terminal.flush()?;
             // Move back to correct cursor location
             if let Some(Loc { x, y }) = self.doc().cursor_loc_in_screen() {
@@ -924,8 +1012,9 @@ impl Editor {
         let file_name = self.prompt("Save as")?;
         self.doc_mut().save_as(&file_name)?;
         if self.doc().file_name.is_none() {
-            let ext = file_name.split('.').last().unwrap();
-            self.highlighter[self.ptr] = self.config
+            let ext = file_name.split('.').last().unwrap_or("");
+            self.highlighter[self.ptr] = self
+                .config
                 .syntax_highlighting
                 .borrow()
                 .get_highlighter(&ext);
