@@ -1,20 +1,20 @@
+mod cli;
 mod config;
 mod editor;
 mod error;
-mod cli;
 mod ui;
 
+use cli::CommandLineInterface;
+use config::{run_key, PLUGIN_BOOTSTRAP, PLUGIN_RUN};
+use editor::Editor;
+use error::Result;
 use kaolinite::event::Event;
 use kaolinite::Loc;
-use error::Result;
-use cli::CommandLineInterface;
-use editor::Editor;
-use ui::Feedback;
-use std::rc::Rc;
-use std::cell::RefCell;
-use config::{PLUGIN_BOOTSTRAP, PLUGIN_RUN};
-use mlua::Lua;
 use mlua::Error::RuntimeError;
+use mlua::Lua;
+use std::cell::RefCell;
+use std::rc::Rc;
+use ui::Feedback;
 
 fn main() {
     // Interact with user to find out what they want to do
@@ -45,8 +45,14 @@ fn run(cli: CommandLineInterface) -> Result<()> {
 
     // Load config and initialise
     lua.load(PLUGIN_BOOTSTRAP).exec()?;
-    editor.borrow_mut().load_config(cli.config_path, &lua).unwrap();
-    editor.borrow_mut().init()?;
+    if editor
+        .borrow_mut()
+        .load_config(cli.config_path, &lua)
+        .is_err()
+    {
+        editor.borrow_mut().feedback =
+            Feedback::Error("Failed to load configuration file".to_string());
+    }
 
     // Open files user has asked to open
     for (c, file) in cli.to_open.iter().enumerate() {
@@ -94,12 +100,13 @@ fn run(cli: CommandLineInterface) -> Result<()> {
     lua.load(PLUGIN_RUN).exec()?;
 
     // Run the editor and handle errors if applicable
+    editor.borrow_mut().init()?;
     while editor.borrow().active {
-        let cycle = editor.borrow_mut().cycle();
+        let cycle = editor.borrow_mut().cycle(&lua);
         match cycle {
             Ok(Some(key)) => {
                 // Form the corresponding lua code to run and run it
-                let code = format!("(event_mapping[\"{key}\"] or error(\"key not bound\"))()");
+                let code = run_key(&key);
                 let result = lua.load(&code).exec();
                 // Check the result
                 match result {
@@ -108,11 +115,11 @@ fn run(cli: CommandLineInterface) -> Result<()> {
                         // Work out if the key has been bound
                         if msg.contains(&"key not bound") {
                             if key.contains(&"_") && key != "_" && !key.starts_with("shift") {
-                                editor.borrow_mut().feedback = 
+                                editor.borrow_mut().feedback =
                                     Feedback::Error(format!("The key binding {key} is not set"));
                             }
                         } else {
-                            let mut msg: String = msg.split("\n").next().unwrap().to_string();
+                            let mut msg: String = msg.split("\n").next().unwrap_or("").to_string();
                             // Work out the type of error and display an appropriate message
                             if msg.starts_with("[") {
                                 msg = msg.split(":").skip(3).collect::<Vec<&str>>().join(":");
@@ -121,11 +128,13 @@ fn run(cli: CommandLineInterface) -> Result<()> {
                                 msg = format!(": {msg}");
                             }
                             // Send out the error
-                            editor.borrow_mut().feedback = Feedback::Error(format!("Lua error occured{msg}"));
+                            editor.borrow_mut().feedback =
+                                Feedback::Error(format!("Lua error occured{msg}"));
                         }
                     }
                     Err(err) => {
-                        editor.borrow_mut().feedback = Feedback::Error(format!("Error occured: {err}"));
+                        editor.borrow_mut().feedback =
+                            Feedback::Error(format!("Error occured: {err}"));
                     }
                     _ => (),
                 }
@@ -156,7 +165,7 @@ fn run_editor_command(editor: &Rc<RefCell<Editor>>, cmd: String, lua: &Lua) {
             let arguments = arguments.join("', '");
             let code = format!("commands['{subcmd}']({{'{arguments}'}})");
             if let Err(err) = lua.load(code).exec() {
-                let line = err.to_string().split("\n").next().unwrap().to_string();
+                let line = err.to_string().split("\n").next().unwrap_or("").to_string();
                 editor.borrow_mut().feedback = Feedback::Error(line);
             }
         }
